/*
 * kgraft_patch_bsc1099306_kvm_intel
 *
 * Fix for CVE-2018-3646 (kvm_intel.ko part), bsc#1099306
 *
 *  Copyright (c) 2018 SUSE
 *  Author: Nicolai Stange <nstange@suse.de>
 *
 *  Based on the original Linux kernel code. Other copyrights apply.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

#if IS_ENABLED(CONFIG_X86_64) && IS_ENABLED(CONFIG_KVM)

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/kallsyms.h>
#include <linux/kvm_host.h>
#include <asm/vmx.h>
#include <linux/highmem.h>
#include <asm/segment.h>
#include <asm/tlbflush.h>
#include "kgr_patch_bsc1099306_kvm_intel.h"
#include "bsc1099306.h"

#if !IS_MODULE(CONFIG_KVM_INTEL)
#error "Live patch supports only CONFIG_KVM_INTEL=m"
#endif

#define KGR_PATCHED_MODULE "kvm_intel"


struct vcpu_vmx;
struct vmcs12;

/* from arch/x86/kvm/vmx.c */
struct vmcs_config {
	int size;
	int order;
	u32 revision_id;
	u32 pin_based_exec_ctrl;
	u32 cpu_based_exec_ctrl;
	u32 cpu_based_2nd_exec_ctrl;
	u32 vmexit_ctrl;
	u32 vmentry_ctrl;
};

static bool *kgr_kvm_rebooting;
static asmlinkage void (*kgr_kvm_spurious_fault)(void);
static int (*kgr_kvm_set_shared_msr)(unsigned slot, u64 value, u64 mask);
static bool (*kgr_kvm_valid_efer)(struct kvm_vcpu *vcpu, u64 efer);
static int (*kgr_kvm_set_dr)(struct kvm_vcpu *vcpu, int dr, unsigned long val);
static void (*kgr_kvm_release_page_dirty)(struct page *page);
static struct page* (*kgr_gfn_to_page)(struct kvm *kvm, gfn_t gfn);
static void (*kgr_kvm_mmu_unload)(struct kvm_vcpu *vcpu);
static int (*kgr_kvm_init_shadow_ept_mmu)(struct kvm_vcpu *vcpu,
					  struct kvm_mmu *context,
					  bool execonly);
static int (*kgr_kvm_set_cr3)(struct kvm_vcpu *vcpu, unsigned long cr3);
static int (*kgr_kvm_mmu_reset_context)(struct kvm_vcpu *vcpu);
static int (*kgr_kvm_emulate_halt)(struct kvm_vcpu *vcpu);

static bool *kgr_enable_ept;
static bool *kgr_enable_shadow_vmcs;
static bool *kgr_flexpriority_enabled;
static bool *kgr_enable_unrestricted_guest;
static int *kgr_ple_gap;
static bool *kgr_enable_vpid;
static u32 *kgr_nested_vmx_procbased_ctls_low;
static u32 *kgr_nested_vmx_procbased_ctls_high;
static u32 *kgr_nested_vmx_secondary_ctls_low;
static u32 *kgr_nested_vmx_secondary_ctls_high;
static u32 *kgr_nested_vmx_pinbased_ctls_low;
static u32 *kgr_nested_vmx_pinbased_ctls_high;
static u32 *kgr_nested_vmx_exit_ctls_low;
static u32 *kgr_nested_vmx_exit_ctls_high;
static u32 *kgr_nested_vmx_entry_ctls_low;
static u32 *kgr_nested_vmx_entry_ctls_high;
static u32 *kgr_nested_vmx_ept_caps;
static struct vmcs_config *kgr_vmcs_config;
static const ulong *kgr_vmx_return;
static void (*kgr_vmcs_writel)(unsigned long field, unsigned long value);
static unsigned long (*kgr_segment_base)(u16 selector);
static int (*kgr_nested_vmx_check_permission)(struct kvm_vcpu *vcpu);
static void (*kgr_nested_vmx_failInvalid)(struct kvm_vcpu *vcpu);
static void (*kgr_skip_emulated_instruction)(struct kvm_vcpu *vcpu);
static void (*kgr_copy_shadow_to_vmcs12)(struct vcpu_vmx *vmx);
static void (*kgr_nested_vmx_failValid)(struct kvm_vcpu *vcpu,
					u32 vm_instruction_error);
static void (*kgr_load_vmcs12_host_state)(struct kvm_vcpu *vcpu,
					  struct vmcs12 *vmcs12);
static void (*kgr_nested_vmx_succeed)(struct kvm_vcpu *vcpu);
static struct vmcs* (*kgr_alloc_vmcs_cpu)(int cpu);
static void (*kgr_vmcs_clear)(struct vmcs *vmcs);
static void (*kgr_vmx_vcpu_put)(struct kvm_vcpu *vcpu);
static void (*kgr_vmx_vcpu_load)(struct kvm_vcpu *vcpu, int cpu);
static void (*kgr_vmx_set_rflags)(struct kvm_vcpu *vcpu, unsigned long rflags);
static enum hrtimer_restart
(*kgr_vmx_preemption_timer_fn)(struct hrtimer *timer);
static int (*kgr_vmx_vm_has_apicv)(struct kvm *kvm);
static void (*kgr_update_exception_bitmap)(struct kvm_vcpu *vcpu);
static void (*kgr_set_cr4_guest_host_mask)(struct vcpu_vmx *vmx);
static void (*kgr_vmx_flush_tlb)(struct kvm_vcpu *vcpu);
static void (*kgr_vmx_set_cr3)(struct kvm_vcpu *vcpu, unsigned long cr3);
static unsigned long (*kgr_nested_ept_get_cr3)(struct kvm_vcpu *vcpu);
static void (*kgr_nested_ept_inject_page_fault)(struct kvm_vcpu *vcpu,
						struct x86_exception *fault);
static void (*kgr_vmx_set_efer)(struct kvm_vcpu *vcpu, u64 efer);
static void (*kgr_vmx_set_cr0)(struct kvm_vcpu *vcpu, unsigned long cr0);
static int (*kgr_vmx_set_cr4)(struct kvm_vcpu *vcpu, unsigned long cr4);
static void (*kgr_vmx_inject_page_fault_nested)(struct kvm_vcpu *vcpu,
						struct x86_exception *fault);

static struct {
	char *name;
	void **addr;
} kgr_funcs[] = {
	{ "kvm:kvm_rebooting", (void *)&kgr_kvm_rebooting },
	{ "kvm:kvm_spurious_fault", (void *)&kgr_kvm_spurious_fault },
	{ "kvm:kvm_set_shared_msr", (void *)&kgr_kvm_set_shared_msr },
	{ "kvm:kvm_valid_efer", (void *)&kgr_kvm_valid_efer },
	{ "kvm:kvm_set_dr", (void *)&kgr_kvm_set_dr },
	{ "kvm:kvm_release_page_dirty", (void *)&kgr_kvm_release_page_dirty },
	{ "kvm:gfn_to_page", (void *)&kgr_gfn_to_page },
	{ "kvm:kvm_mmu_unload", (void *)&kgr_kvm_mmu_unload },
	{ "kvm:kvm_init_shadow_ept_mmu", (void *)&kgr_kvm_init_shadow_ept_mmu },
	{ "kvm:kvm_set_cr3", (void *)&kgr_kvm_set_cr3 },
	{ "kvm:kvm_mmu_reset_context", (void *)&kgr_kvm_mmu_reset_context },
	{ "kvm:kvm_emulate_halt", (void *)&kgr_kvm_emulate_halt },

	{ "kvm_intel:enable_ept", (void *)&kgr_enable_ept },
	{ "kvm_intel:enable_shadow_vmcs", (void *)&kgr_enable_shadow_vmcs },
	{ "kvm_intel:flexpriority_enabled", (void *)&kgr_flexpriority_enabled },
	{ "kvm_intel:enable_unrestricted_guest",
				(void *)&kgr_enable_unrestricted_guest },
	{ "kvm_intel:ple_gap", (void *)&kgr_ple_gap },
	{ "kvm_intel:enable_vpid", (void *)&kgr_enable_vpid },
	{ "kvm_intel:nested_vmx_procbased_ctls_low",
				(void *)&kgr_nested_vmx_procbased_ctls_low },
	{ "kvm_intel:nested_vmx_procbased_ctls_high",
				(void *)&kgr_nested_vmx_procbased_ctls_high },
	{ "kvm_intel:nested_vmx_secondary_ctls_low",
				(void *)&kgr_nested_vmx_secondary_ctls_low },
	{ "kvm_intel:nested_vmx_secondary_ctls_high",
				(void *)&kgr_nested_vmx_secondary_ctls_high },
	{ "kvm_intel:nested_vmx_pinbased_ctls_low",
				(void *)&kgr_nested_vmx_pinbased_ctls_low },
	{ "kvm_intel:nested_vmx_pinbased_ctls_high",
				(void *)&kgr_nested_vmx_pinbased_ctls_high },
	{ "kvm_intel:nested_vmx_exit_ctls_low",
				(void *)&kgr_nested_vmx_exit_ctls_low },
	{ "kvm_intel:nested_vmx_exit_ctls_high",
				(void *)&kgr_nested_vmx_exit_ctls_high },
	{ "kvm_intel:nested_vmx_entry_ctls_low",
				(void *)&kgr_nested_vmx_entry_ctls_low },
	{ "kvm_intel:nested_vmx_entry_ctls_high",
				(void *)&kgr_nested_vmx_entry_ctls_high },
	{ "kvm_intel:nested_vmx_ept_caps",
				(void *)&kgr_nested_vmx_ept_caps },
	{ "kvm_intel:vmcs_config", (void *)&kgr_vmcs_config },
	{ "kvm_intel:vmx_return", (void *)&kgr_vmx_return },
	{ "kvm_intel:vmcs_writel", (void *)&kgr_vmcs_writel },
	{ "kvm_intel:segment_base", (void *)&kgr_segment_base },
	{ "kvm_intel:nested_vmx_check_permission",
				(void *)&kgr_nested_vmx_check_permission },
	{ "kvm_intel:nested_vmx_failInvalid",
				(void *)&kgr_nested_vmx_failInvalid },
	{ "kvm_intel:skip_emulated_instruction",
				(void *)&kgr_skip_emulated_instruction },
	{ "kvm_intel:copy_shadow_to_vmcs12",
				(void *)&kgr_copy_shadow_to_vmcs12 },
	{ "kvm_intel:nested_vmx_failValid",
				(void *)&kgr_nested_vmx_failValid },
	{ "kvm_intel:load_vmcs12_host_state",
				(void *)&kgr_load_vmcs12_host_state },
	{ "kvm_intel:nested_vmx_succeed",
				(void *)&kgr_nested_vmx_succeed },
	{ "kvm_intel:alloc_vmcs_cpu", (void *)&kgr_alloc_vmcs_cpu },
	{ "kvm_intel:vmcs_clear", (void *)&kgr_vmcs_clear },
	{ "kvm_intel:vmx_vcpu_put", (void *)&kgr_vmx_vcpu_put },
	{ "kvm_intel:vmx_vcpu_load", (void *)&kgr_vmx_vcpu_load },
	{ "kvm_intel:vmx_set_rflags", (void *)&kgr_vmx_set_rflags },
	{ "kvm_intel:vmx_preemption_timer_fn",
				(void *)&kgr_vmx_preemption_timer_fn },
	{ "kvm_intel:vmx_vm_has_apicv", (void *)&kgr_vmx_vm_has_apicv },
	{ "kvm_intel:update_exception_bitmap",
				(void *)&kgr_update_exception_bitmap },
	{ "kvm_intel:set_cr4_guest_host_mask",
				(void *)&kgr_set_cr4_guest_host_mask },
	{ "kvm_intel:vmx_flush_tlb", (void *)&kgr_vmx_flush_tlb },
	{ "kvm_intel:vmx_set_cr3", (void *)&kgr_vmx_set_cr3 },
	{ "kvm_intel:nested_ept_get_cr3", (void *)&kgr_nested_ept_get_cr3 },
	{ "kvm_intel:nested_ept_inject_page_fault",
				(void *)&kgr_nested_ept_inject_page_fault },
	{ "kvm_intel:vmx_set_efer", (void *)&kgr_vmx_set_efer },
	{ "kvm_intel:vmx_set_cr0", (void *)&kgr_vmx_set_cr0 },
	{ "kvm_intel:vmx_set_cr4", (void *)&kgr_vmx_set_cr4 },
	{ "kvm_intel:vmx_inject_page_fault_nested",
				(void *)&kgr_vmx_inject_page_fault_nested },
};


/* from arch/x86/include/asm/kvm_host.h */
/*
 * Avoid module dependency on kvm.ko.
 *
 * This is the original macro but with the load from kvm_rebooting
 * replaced by an indirect load from *kgr_kvm_rebooting and the call
 * to kvm_spurious_fault() replaced by an indirect call to
 * *kgr_kvm_spurious_fault.
 */
#define kgr____kvm_handle_fault_on_reboot(insn, cleanup_insn)	\
	"666: " insn "\n\t"				\
	"668: \n\t"					\
	".pushsection .fixup, \"ax\" \n"		\
	"667: \n\t"					\
	cleanup_insn "\n\t"				\
	"pushq %%rax \n\t"				\
	"movq kgr_kvm_rebooting, %%rax\n\t"		\
	"cmpb $0, (%%rax) \n\t"			\
	"popq %%rax \n\t"				\
	"jne 668b \n\t"				\
	__ASM_SIZE(push) " $666b \n\t"			\
	"movq kgr_kvm_spurious_fault, %%rax \n\t"	\
	"call *%%rax \n\t"				\
	".popsection \n\t"				\
	_ASM_EXTABLE(666b, 667b)


/* from arch/x86/kvm/x86.h */
/* inlined */
static inline int kgr_is_long_mode(struct kvm_vcpu *vcpu)
{
#ifdef CONFIG_X86_64
	return vcpu->arch.efer & EFER_LMA;
#else
	return 0;
#endif
}


/* from arch/x86/kvm/irq.h */
/* inlined */
static inline struct kvm_pic *kgr_pic_irqchip(struct kvm *kvm)
{
	return kvm->arch.vpic;
}

/* inlined */
static inline int kgr_irqchip_in_kernel(struct kvm *kvm)
{
	int ret;

	ret = (kgr_pic_irqchip(kvm) != NULL);
	smp_rmb();
	return ret;
}


/* from arch/x86/kvm/kvm_cache_regs.h */
/* inlined */
static inline void kgr_kvm_register_write(struct kvm_vcpu *vcpu,
					  enum kvm_reg reg,
					  unsigned long val)
{
	vcpu->arch.regs[reg] = val;
	__set_bit(reg, (unsigned long *)&vcpu->arch.regs_dirty);
	__set_bit(reg, (unsigned long *)&vcpu->arch.regs_avail);
}

/* inlined */
static inline void kgr_enter_guest_mode(struct kvm_vcpu *vcpu)
{
	vcpu->arch.hflags |= HF_GUEST_MASK;
}


/* from arch/x86/kvm/vmx.c */
#define kgr__ex_clear(x, reg) \
	kgr____kvm_handle_fault_on_reboot(x, "xor " reg " , " reg)

#define KGR_VMX_MISC_EMULATED_PREEMPTION_TIMER_RATE 5

#define KGR_NR_AUTOLOAD_MSRS 8
#define KGR_VMCS02_POOL_SIZE 1

struct loaded_vmcs {
	struct vmcs *vmcs;
	int cpu;
	int launched;
	struct list_head loaded_vmcss_on_cpu_link;
};

struct shared_msr_entry {
	unsigned index;
	u64 data;
	u64 mask;
};

typedef u64 natural_width;
struct __packed vmcs12 {
	/* According to the Intel spec, a VMCS region must start with the
	 * following two fields. Then follow implementation-specific data.
	 */
	u32 revision_id;
	u32 abort;

	u32 launch_state; /* set to 0 by VMCLEAR, to 1 by VMLAUNCH */
	u32 padding[7]; /* room for future expansion */

	u64 io_bitmap_a;
	u64 io_bitmap_b;
	u64 msr_bitmap;
	u64 vm_exit_msr_store_addr;
	u64 vm_exit_msr_load_addr;
	u64 vm_entry_msr_load_addr;
	u64 tsc_offset;
	u64 virtual_apic_page_addr;
	u64 apic_access_addr;
	u64 ept_pointer;
	u64 guest_physical_address;
	u64 vmcs_link_pointer;
	u64 guest_ia32_debugctl;
	u64 guest_ia32_pat;
	u64 guest_ia32_efer;
	u64 guest_ia32_perf_global_ctrl;
	u64 guest_pdptr0;
	u64 guest_pdptr1;
	u64 guest_pdptr2;
	u64 guest_pdptr3;
	u64 host_ia32_pat;
	u64 host_ia32_efer;
	u64 host_ia32_perf_global_ctrl;
	u64 padding64[8]; /* room for future expansion */
	/*
	 * To allow migration of L1 (complete with its L2 guests) between
	 * machines of different natural widths (32 or 64 bit), we cannot have
	 * unsigned long fields with no explict size. We use u64 (aliased
	 * natural_width) instead. Luckily, x86 is little-endian.
	 */
	natural_width cr0_guest_host_mask;
	natural_width cr4_guest_host_mask;
	natural_width cr0_read_shadow;
	natural_width cr4_read_shadow;
	natural_width cr3_target_value0;
	natural_width cr3_target_value1;
	natural_width cr3_target_value2;
	natural_width cr3_target_value3;
	natural_width exit_qualification;
	natural_width guest_linear_address;
	natural_width guest_cr0;
	natural_width guest_cr3;
	natural_width guest_cr4;
	natural_width guest_es_base;
	natural_width guest_cs_base;
	natural_width guest_ss_base;
	natural_width guest_ds_base;
	natural_width guest_fs_base;
	natural_width guest_gs_base;
	natural_width guest_ldtr_base;
	natural_width guest_tr_base;
	natural_width guest_gdtr_base;
	natural_width guest_idtr_base;
	natural_width guest_dr7;
	natural_width guest_rsp;
	natural_width guest_rip;
	natural_width guest_rflags;
	natural_width guest_pending_dbg_exceptions;
	natural_width guest_sysenter_esp;
	natural_width guest_sysenter_eip;
	natural_width host_cr0;
	natural_width host_cr3;
	natural_width host_cr4;
	natural_width host_fs_base;
	natural_width host_gs_base;
	natural_width host_tr_base;
	natural_width host_gdtr_base;
	natural_width host_idtr_base;
	natural_width host_ia32_sysenter_esp;
	natural_width host_ia32_sysenter_eip;
	natural_width host_rsp;
	natural_width host_rip;
	natural_width paddingl[8]; /* room for future expansion */
	u32 pin_based_vm_exec_control;
	u32 cpu_based_vm_exec_control;
	u32 exception_bitmap;
	u32 page_fault_error_code_mask;
	u32 page_fault_error_code_match;
	u32 cr3_target_count;
	u32 vm_exit_controls;
	u32 vm_exit_msr_store_count;
	u32 vm_exit_msr_load_count;
	u32 vm_entry_controls;
	u32 vm_entry_msr_load_count;
	u32 vm_entry_intr_info_field;
	u32 vm_entry_exception_error_code;
	u32 vm_entry_instruction_len;
	u32 tpr_threshold;
	u32 secondary_vm_exec_control;
	u32 vm_instruction_error;
	u32 vm_exit_reason;
	u32 vm_exit_intr_info;
	u32 vm_exit_intr_error_code;
	u32 idt_vectoring_info_field;
	u32 idt_vectoring_error_code;
	u32 vm_exit_instruction_len;
	u32 vmx_instruction_info;
	u32 guest_es_limit;
	u32 guest_cs_limit;
	u32 guest_ss_limit;
	u32 guest_ds_limit;
	u32 guest_fs_limit;
	u32 guest_gs_limit;
	u32 guest_ldtr_limit;
	u32 guest_tr_limit;
	u32 guest_gdtr_limit;
	u32 guest_idtr_limit;
	u32 guest_es_ar_bytes;
	u32 guest_cs_ar_bytes;
	u32 guest_ss_ar_bytes;
	u32 guest_ds_ar_bytes;
	u32 guest_fs_ar_bytes;
	u32 guest_gs_ar_bytes;
	u32 guest_ldtr_ar_bytes;
	u32 guest_tr_ar_bytes;
	u32 guest_interruptibility_info;
	u32 guest_activity_state;
	u32 guest_sysenter_cs;
	u32 host_ia32_sysenter_cs;
	u32 vmx_preemption_timer_value;
	u32 padding32[7]; /* room for future expansion */
	u16 virtual_processor_id;
	u16 guest_es_selector;
	u16 guest_cs_selector;
	u16 guest_ss_selector;
	u16 guest_ds_selector;
	u16 guest_fs_selector;
	u16 guest_gs_selector;
	u16 guest_ldtr_selector;
	u16 guest_tr_selector;
	u16 host_es_selector;
	u16 host_cs_selector;
	u16 host_ss_selector;
	u16 host_ds_selector;
	u16 host_fs_selector;
	u16 host_gs_selector;
	u16 host_tr_selector;
};

struct vmcs02_list {
	struct list_head list;
	gpa_t vmptr;
	struct loaded_vmcs vmcs02;
};

struct nested_vmx {
	/* Has the level1 guest done vmxon? */
	bool vmxon;

	/* The guest-physical address of the current VMCS L1 keeps for L2 */
	gpa_t current_vmptr;
	/* The host-usable pointer to the above */
	struct page *current_vmcs12_page;
	struct vmcs12 *current_vmcs12;
	struct vmcs *current_shadow_vmcs;
	/*
	 * Indicates if the shadow vmcs must be updated with the
	 * data hold by vmcs12
	 */
	bool sync_shadow_vmcs;

	/* vmcs02_list cache of VMCSs recently used to run L2 guests */
	struct list_head vmcs02_pool;
	int vmcs02_num;
	u64 vmcs01_tsc_offset;
	/* L2 must run next, and mustn't decide to exit to L1. */
	bool nested_run_pending;
	/*
	 * Guest pages referred to in vmcs02 with host-physical pointers, so
	 * we must keep them pinned while L2 runs.
	 */
	struct page *apic_access_page;
	u64 msr_ia32_feature_control;

	struct hrtimer preemption_timer;
	bool preemption_timer_expired;
};

struct pi_desc {
	u32 pir[8];     /* Posted interrupt requested */
	u32 control;	/* bit 0 of control is outstanding notification bit */
	u32 rsvd[7];
} __aligned(64);

struct vcpu_vmx {
	struct kvm_vcpu       vcpu;
	unsigned long         host_rsp;
	u8                    fail;
	u8                    cpl;
	bool                  nmi_known_unmasked;
	u32                   exit_intr_info;
	u32                   idt_vectoring_info;
	ulong                 rflags;
	struct shared_msr_entry *guest_msrs;
	int                   nmsrs;
	int                   save_nmsrs;
	unsigned long	      host_idt_base;
#ifdef CONFIG_X86_64
	u64 		      msr_host_kernel_gs_base;
	u64 		      msr_guest_kernel_gs_base;
#endif
	u32 vm_entry_controls_shadow;
	u32 vm_exit_controls_shadow;
	/*
	 * loaded_vmcs points to the VMCS currently used in this vcpu. For a
	 * non-nested (L1) guest, it always points to vmcs01. For a nested
	 * guest (L2), it points to a different VMCS.
	 */
	struct loaded_vmcs    vmcs01;
	struct loaded_vmcs   *loaded_vmcs;
	bool                  __launched; /* temporary, used in vmx_vcpu_run */
	struct msr_autoload {
		unsigned nr;
		struct vmx_msr_entry guest[KGR_NR_AUTOLOAD_MSRS];
		struct vmx_msr_entry host[KGR_NR_AUTOLOAD_MSRS];
	} msr_autoload;
	struct {
		int           loaded;
		u16           fs_sel, gs_sel, ldt_sel;
#ifdef CONFIG_X86_64
		u16           ds_sel, es_sel;
#endif
		int           gs_ldt_reload_needed;
		int           fs_reload_needed;
		unsigned long vmcs_host_cr4;    /* May not match real cr4 */
	} host_state;
	struct {
		int vm86_active;
		ulong save_rflags;
		struct kvm_segment segs[8];
	} rmode;
	struct {
		u32 bitmask; /* 4 bits per segment (1 bit per field) */
		struct kvm_save_segment {
			u16 selector;
			unsigned long base;
			u32 limit;
			u32 ar;
		} seg[8];
	} segment_cache;
	int vpid;
	bool emulation_required;

	/* Support for vnmi-less CPUs */
	int soft_vnmi_blocked;
	ktime_t entry_time;
	s64 vnmi_blocked_time;
	u32 exit_reason;

	bool rdtscp_enabled;

	/* Posted interrupt descriptor */
	struct pi_desc pi_desc;

	/* Support for a guest hypervisor (nested VMX) */
	struct nested_vmx nested;
};

/* inlined */
static inline struct vcpu_vmx *kgr_to_vmx(struct kvm_vcpu *vcpu)
{
	return container_of(vcpu, struct vcpu_vmx, vcpu);
}

/* inlined */
static inline struct vmcs12 *kgr_get_vmcs12(struct kvm_vcpu *vcpu)
{
	return kgr_to_vmx(vcpu)->nested.current_vmcs12;
}

/* inlined */
static struct page *kgr_nested_get_page(struct kvm_vcpu *vcpu, gpa_t addr)
{
	struct page *page = kgr_gfn_to_page(vcpu->kvm, addr >> PAGE_SHIFT);
	if (is_error_page(page))
		return NULL;

	return page;
}

/* inlined */
/* ine 720 */
static void kgr_nested_release_page(struct page *page)
{
	kgr_kvm_release_page_dirty(page);
}

/* inlined */
static inline bool kgr_cpu_has_vmx_tpr_shadow(void)
{
	return kgr_vmcs_config->cpu_based_exec_ctrl & CPU_BASED_TPR_SHADOW;
}

/* inlined */
static inline bool kgr_vm_need_tpr_shadow(struct kvm *kvm)
{
	return (kgr_cpu_has_vmx_tpr_shadow()) && (kgr_irqchip_in_kernel(kvm));
}

/* inlined */
static inline bool kgr_cpu_has_secondary_exec_ctrls(void)
{
	return kgr_vmcs_config->cpu_based_exec_ctrl &
		CPU_BASED_ACTIVATE_SECONDARY_CONTROLS;
}

/* inlined */
static inline bool kgr_vm_need_virtualize_apic_accesses(struct kvm *kvm)
{
	return *kgr_flexpriority_enabled && kgr_irqchip_in_kernel(kvm);
}

/* inlined */
static inline bool kgr_nested_cpu_has(struct vmcs12 *vmcs12, u32 bit)
{
	return vmcs12->cpu_based_vm_exec_control & bit;
}

/* inlined */
static inline bool kgr_nested_cpu_has2(struct vmcs12 *vmcs12, u32 bit)
{
	return (vmcs12->cpu_based_vm_exec_control &
			CPU_BASED_ACTIVATE_SECONDARY_CONTROLS) &&
		(vmcs12->secondary_vm_exec_control & bit);
}

/* inlined */
static inline bool kgr_nested_cpu_has_preemption_timer(struct vmcs12 *vmcs12)
{
	return vmcs12->pin_based_vm_exec_control &
		PIN_BASED_VMX_PREEMPTION_TIMER;
}

/* inlined */
static inline int kgr_nested_cpu_has_ept(struct vmcs12 *vmcs12)
{
	return kgr_nested_cpu_has2(vmcs12, SECONDARY_EXEC_ENABLE_EPT);
}

/* inlined */
static inline void kgr_loaded_vmcs_init(struct loaded_vmcs *loaded_vmcs)
{
	kgr_vmcs_clear(loaded_vmcs->vmcs);
	loaded_vmcs->cpu = -1;
	loaded_vmcs->launched = 0;
}

/* inlined */
static __always_inline unsigned long kgr_vmcs_readl(unsigned long field)
{
	unsigned long value;

	asm volatile (kgr__ex_clear(ASM_VMX_VMREAD_RDX_RAX, "%0")
		      : "=a"(value) : "d"(field) : "cc");
	return value;
}

/* inlined */
static __always_inline u32 kgr_vmcs_read32(unsigned long field)
{
	return kgr_vmcs_readl(field);
}

/* inlined */
static __always_inline u64 kgr_vmcs_read64(unsigned long field)
{
#ifdef CONFIG_X86_64
	return kgr_vmcs_readl(field);
#else
	return kgr_vmcs_readl(field) | ((u64)kgr_vmcs_readl(field+1) << 32);
#endif
}

/* inlined */
static void kgr_vmcs_write16(unsigned long field, u16 value)
{
	kgr_vmcs_writel(field, value);
}

/* inlined */
static void kgr_vmcs_write32(unsigned long field, u32 value)
{
	kgr_vmcs_writel(field, value);
}

/* inlined */
static void kgr_vmcs_write64(unsigned long field, u64 value)
{
	kgr_vmcs_writel(field, value);
#ifndef CONFIG_X86_64
	asm volatile ("");
	kgr_vmcs_writel(field+1, value >> 32);
#endif
}

/* inlined */
static inline void kgr_vm_entry_controls_init(struct vcpu_vmx *vmx, u32 val)
{
	kgr_vmcs_write32(VM_ENTRY_CONTROLS, val);
	vmx->vm_entry_controls_shadow = val;
}

/* inlined */
static void kgr_vmx_segment_cache_clear(struct vcpu_vmx *vmx)
{
	vmx->segment_cache.bitmask = 0;
}

/* inlined */
static inline unsigned long kgr_nested_read_cr0(struct vmcs12 *fields)
{
	return (fields->guest_cr0 & ~fields->cr0_guest_host_mask) |
		(fields->cr0_read_shadow & fields->cr0_guest_host_mask);
}

/* inlined */
static inline unsigned long kgr_nested_read_cr4(struct vmcs12 *fields)
{
	return (fields->guest_cr4 & ~fields->cr4_guest_host_mask) |
		(fields->cr4_read_shadow & fields->cr4_guest_host_mask);
}

/* inlined */
static inline bool kgr_vmx_control_verify(u32 control, u32 low, u32 high)
{
	/*
	 * Bits 0 in high must be 0, and bits 1 in low must be 1.
	 */
	return ((control & high) | low) == control;
}

#define KGR_VMXON_CR0_ALWAYSON	(X86_CR0_PE | X86_CR0_PG | X86_CR0_NE)
#define KGR_VMXON_CR4_ALWAYSON	X86_CR4_VMXE

/* inlined */
static struct vmcs *kgr_alloc_vmcs(void)
{
	return kgr_alloc_vmcs_cpu(raw_smp_processor_id());
}

/* optimized */
static void kgr_vmx_set_constant_host_state(struct vcpu_vmx *vmx)
{
	u32 low32, high32;
	unsigned long tmpl;
	struct desc_ptr dt;
	unsigned long cr4;

	kgr_vmcs_writel(HOST_CR0, read_cr0() & ~X86_CR0_TS);  /* 22.2.3 */
	kgr_vmcs_writel(HOST_CR3, read_cr3());  /* 22.2.3  FIXME: shadow tables */

	/* Save the most likely value for this task's CR4 in the VMCS. */
	cr4 = read_cr4();
	kgr_vmcs_writel(HOST_CR4, cr4);			/* 22.2.3, 22.2.5 */
	vmx->host_state.vmcs_host_cr4 = cr4;

	kgr_vmcs_write16(HOST_CS_SELECTOR, __KERNEL_CS);  /* 22.2.4 */
#ifdef CONFIG_X86_64
	/*
	 * Load null selectors, so we can avoid reloading them in
	 * __vmx_load_host_state(), in case userspace uses the null selectors
	 * too (the expected case).
	 */
	kgr_vmcs_write16(HOST_DS_SELECTOR, 0);
	kgr_vmcs_write16(HOST_ES_SELECTOR, 0);
#else
	kgr_vmcs_write16(HOST_DS_SELECTOR, __KERNEL_DS);  /* 22.2.4 */
	kgr_vmcs_write16(HOST_ES_SELECTOR, __KERNEL_DS);  /* 22.2.4 */
#endif
	kgr_vmcs_write16(HOST_SS_SELECTOR, __KERNEL_DS);  /* 22.2.4 */
	kgr_vmcs_write16(HOST_TR_SELECTOR, GDT_ENTRY_TSS*8);  /* 22.2.4 */

	native_store_idt(&dt);
	kgr_vmcs_writel(HOST_IDTR_BASE, dt.address);   /* 22.2.4 */
	vmx->host_idt_base = dt.address;

	kgr_vmcs_writel(HOST_RIP, *kgr_vmx_return); /* 22.2.5 */

	rdmsr(MSR_IA32_SYSENTER_CS, low32, high32);
	kgr_vmcs_write32(HOST_IA32_SYSENTER_CS, low32);
	rdmsrl(MSR_IA32_SYSENTER_EIP, tmpl);
	kgr_vmcs_writel(HOST_IA32_SYSENTER_EIP, tmpl);   /* 22.2.3 */

	if (kgr_vmcs_config->vmexit_ctrl & VM_EXIT_LOAD_IA32_PAT) {
		rdmsr(MSR_IA32_CR_PAT, low32, high32);
		kgr_vmcs_write64(HOST_IA32_PAT, low32 | ((u64) high32 << 32));
	}
}

/* optimized */
static u32 kgr_vmx_exec_control(struct vcpu_vmx *vmx)
{
	u32 exec_control = kgr_vmcs_config->cpu_based_exec_ctrl;
	if (!kgr_vm_need_tpr_shadow(vmx->vcpu.kvm)) {
		exec_control &= ~CPU_BASED_TPR_SHADOW;
#ifdef CONFIG_X86_64
		exec_control |= CPU_BASED_CR8_STORE_EXITING |
				CPU_BASED_CR8_LOAD_EXITING;
#endif
	}
	if (!*kgr_enable_ept)
		exec_control |= CPU_BASED_CR3_STORE_EXITING |
				CPU_BASED_CR3_LOAD_EXITING  |
				CPU_BASED_INVLPG_EXITING;
	return exec_control;
}

/* optimized */
static u32 kgr_vmx_secondary_exec_control(struct vcpu_vmx *vmx)
{
	u32 exec_control = kgr_vmcs_config->cpu_based_2nd_exec_ctrl;
	if (!kgr_vm_need_virtualize_apic_accesses(vmx->vcpu.kvm))
		exec_control &= ~SECONDARY_EXEC_VIRTUALIZE_APIC_ACCESSES;
	if (vmx->vpid == 0)
		exec_control &= ~SECONDARY_EXEC_ENABLE_VPID;
	if (!*kgr_enable_ept) {
		exec_control &= ~SECONDARY_EXEC_ENABLE_EPT;
		*kgr_enable_unrestricted_guest = 0;
		/* Enable INVPCID for non-ept guests may cause performance regression. */
		exec_control &= ~SECONDARY_EXEC_ENABLE_INVPCID;
	}
	if (!*kgr_enable_unrestricted_guest)
		exec_control &= ~SECONDARY_EXEC_UNRESTRICTED_GUEST;
	if (!*kgr_ple_gap)
		exec_control &= ~SECONDARY_EXEC_PAUSE_LOOP_EXITING;
	if (!kgr_vmx_vm_has_apicv(vmx->vcpu.kvm))
		exec_control &= ~(SECONDARY_EXEC_APIC_REGISTER_VIRT |
				  SECONDARY_EXEC_VIRTUAL_INTR_DELIVERY);
	exec_control &= ~SECONDARY_EXEC_VIRTUALIZE_X2APIC_MODE;
	/* SECONDARY_EXEC_SHADOW_VMCS is enabled when L1 executes VMPTRLD
	   (handle_vmptrld).
	   We can NOT enable shadow_vmcs here because we don't have yet
	   a current VMCS12
	*/
	exec_control &= ~SECONDARY_EXEC_SHADOW_VMCS;
	return exec_control;
}

/* inlined */
static bool kgr_nested_cr0_valid(struct vmcs12 *vmcs12, unsigned long val)
{
	unsigned long always_on = KGR_VMXON_CR0_ALWAYSON;

	if (*kgr_nested_vmx_secondary_ctls_high &
		SECONDARY_EXEC_UNRESTRICTED_GUEST &&
	    kgr_nested_cpu_has2(vmcs12, SECONDARY_EXEC_UNRESTRICTED_GUEST))
		always_on &= ~(X86_CR0_PE | X86_CR0_PG);
	return (val & always_on) == always_on;
}

/* inlined */
static struct loaded_vmcs *kgr_nested_get_current_vmcs02(struct vcpu_vmx *vmx)
{
	struct vmcs02_list *item;
	list_for_each_entry(item, &vmx->nested.vmcs02_pool, list)
		if (item->vmptr == vmx->nested.current_vmptr) {
			list_move(&item->list, &vmx->nested.vmcs02_pool);
			return &item->vmcs02;
		}

	if (vmx->nested.vmcs02_num >= max(KGR_VMCS02_POOL_SIZE, 1)) {
		/* Recycle the least recently used VMCS. */
		item = list_entry(vmx->nested.vmcs02_pool.prev,
			struct vmcs02_list, list);
		item->vmptr = vmx->nested.current_vmptr;
		list_move(&item->list, &vmx->nested.vmcs02_pool);
		return &item->vmcs02;
	}

	/* Create a new VMCS */
	item = kmalloc(sizeof(struct vmcs02_list), GFP_KERNEL);
	if (!item)
		return NULL;
	item->vmcs02.vmcs = kgr_alloc_vmcs();
	if (!item->vmcs02.vmcs) {
		kfree(item);
		return NULL;
	}
	kgr_loaded_vmcs_init(&item->vmcs02);
	item->vmptr = vmx->nested.current_vmptr;
	list_add(&(item->list), &(vmx->nested.vmcs02_pool));
	vmx->nested.vmcs02_num++;
	return &item->vmcs02;
}

/* inlined */
static int kgr_nested_vmx_check_vmcs12(struct kvm_vcpu *vcpu)
{
	struct vcpu_vmx *vmx = kgr_to_vmx(vcpu);
	if (vmx->nested.current_vmptr == -1ull) {
		kgr_nested_vmx_failInvalid(vcpu);
		kgr_skip_emulated_instruction(vcpu);
		return 0;
	}
	return 1;
}

/* inlined */
static int kgr_nested_ept_init_mmu_context(struct kvm_vcpu *vcpu)
{
	int r = kgr_kvm_init_shadow_ept_mmu(vcpu, &vcpu->arch.mmu,
			*kgr_nested_vmx_ept_caps & VMX_EPT_EXECUTE_ONLY_BIT);

	vcpu->arch.mmu.set_cr3           = kgr_vmx_set_cr3;
	vcpu->arch.mmu.get_cr3           = kgr_nested_ept_get_cr3;
	vcpu->arch.mmu.inject_page_fault = kgr_nested_ept_inject_page_fault;

	vcpu->arch.walk_mmu              = &vcpu->arch.nested_mmu;

	return r;
}

/* inlined */
static void kgr_vmx_start_preemption_timer(struct kvm_vcpu *vcpu)
{
	u64 preemption_timeout = kgr_get_vmcs12(vcpu)->vmx_preemption_timer_value;
	struct vcpu_vmx *vmx = kgr_to_vmx(vcpu);

	if (vcpu->arch.virtual_tsc_khz == 0)
		return;

	/* Make sure short timeouts reliably trigger an immediate vmexit.
	 * hrtimer_start does not guarantee this. */
	if (preemption_timeout <= 1) {
		kgr_vmx_preemption_timer_fn(&vmx->nested.preemption_timer);
		return;
	}

	preemption_timeout <<= KGR_VMX_MISC_EMULATED_PREEMPTION_TIMER_RATE;
	preemption_timeout *= 1000000;
	do_div(preemption_timeout, vcpu->arch.virtual_tsc_khz);
	hrtimer_start(&vmx->nested.preemption_timer,
		      ns_to_ktime(preemption_timeout), HRTIMER_MODE_REL);
}

/* optimized */
static void kgr_prepare_vmcs02(struct kvm_vcpu *vcpu, struct vmcs12 *vmcs12)
{
	struct vcpu_vmx *vmx = kgr_to_vmx(vcpu);
	u32 exec_control;

	kgr_vmcs_write16(GUEST_ES_SELECTOR, vmcs12->guest_es_selector);
	kgr_vmcs_write16(GUEST_CS_SELECTOR, vmcs12->guest_cs_selector);
	kgr_vmcs_write16(GUEST_SS_SELECTOR, vmcs12->guest_ss_selector);
	kgr_vmcs_write16(GUEST_DS_SELECTOR, vmcs12->guest_ds_selector);
	kgr_vmcs_write16(GUEST_FS_SELECTOR, vmcs12->guest_fs_selector);
	kgr_vmcs_write16(GUEST_GS_SELECTOR, vmcs12->guest_gs_selector);
	kgr_vmcs_write16(GUEST_LDTR_SELECTOR, vmcs12->guest_ldtr_selector);
	kgr_vmcs_write16(GUEST_TR_SELECTOR, vmcs12->guest_tr_selector);
	kgr_vmcs_write32(GUEST_ES_LIMIT, vmcs12->guest_es_limit);
	kgr_vmcs_write32(GUEST_CS_LIMIT, vmcs12->guest_cs_limit);
	kgr_vmcs_write32(GUEST_SS_LIMIT, vmcs12->guest_ss_limit);
	kgr_vmcs_write32(GUEST_DS_LIMIT, vmcs12->guest_ds_limit);
	kgr_vmcs_write32(GUEST_FS_LIMIT, vmcs12->guest_fs_limit);
	kgr_vmcs_write32(GUEST_GS_LIMIT, vmcs12->guest_gs_limit);
	kgr_vmcs_write32(GUEST_LDTR_LIMIT, vmcs12->guest_ldtr_limit);
	kgr_vmcs_write32(GUEST_TR_LIMIT, vmcs12->guest_tr_limit);
	kgr_vmcs_write32(GUEST_GDTR_LIMIT, vmcs12->guest_gdtr_limit);
	kgr_vmcs_write32(GUEST_IDTR_LIMIT, vmcs12->guest_idtr_limit);
	kgr_vmcs_write32(GUEST_ES_AR_BYTES, vmcs12->guest_es_ar_bytes);
	kgr_vmcs_write32(GUEST_CS_AR_BYTES, vmcs12->guest_cs_ar_bytes);
	kgr_vmcs_write32(GUEST_SS_AR_BYTES, vmcs12->guest_ss_ar_bytes);
	kgr_vmcs_write32(GUEST_DS_AR_BYTES, vmcs12->guest_ds_ar_bytes);
	kgr_vmcs_write32(GUEST_FS_AR_BYTES, vmcs12->guest_fs_ar_bytes);
	kgr_vmcs_write32(GUEST_GS_AR_BYTES, vmcs12->guest_gs_ar_bytes);
	kgr_vmcs_write32(GUEST_LDTR_AR_BYTES, vmcs12->guest_ldtr_ar_bytes);
	kgr_vmcs_write32(GUEST_TR_AR_BYTES, vmcs12->guest_tr_ar_bytes);
	kgr_vmcs_writel(GUEST_ES_BASE, vmcs12->guest_es_base);
	kgr_vmcs_writel(GUEST_CS_BASE, vmcs12->guest_cs_base);
	kgr_vmcs_writel(GUEST_SS_BASE, vmcs12->guest_ss_base);
	kgr_vmcs_writel(GUEST_DS_BASE, vmcs12->guest_ds_base);
	kgr_vmcs_writel(GUEST_FS_BASE, vmcs12->guest_fs_base);
	kgr_vmcs_writel(GUEST_GS_BASE, vmcs12->guest_gs_base);
	kgr_vmcs_writel(GUEST_LDTR_BASE, vmcs12->guest_ldtr_base);
	kgr_vmcs_writel(GUEST_TR_BASE, vmcs12->guest_tr_base);
	kgr_vmcs_writel(GUEST_GDTR_BASE, vmcs12->guest_gdtr_base);
	kgr_vmcs_writel(GUEST_IDTR_BASE, vmcs12->guest_idtr_base);

	kgr_vmcs_write64(GUEST_IA32_DEBUGCTL, vmcs12->guest_ia32_debugctl);
	kgr_vmcs_write32(VM_ENTRY_INTR_INFO_FIELD,
		vmcs12->vm_entry_intr_info_field);
	kgr_vmcs_write32(VM_ENTRY_EXCEPTION_ERROR_CODE,
		vmcs12->vm_entry_exception_error_code);
	kgr_vmcs_write32(VM_ENTRY_INSTRUCTION_LEN,
		vmcs12->vm_entry_instruction_len);
	kgr_vmcs_write32(GUEST_INTERRUPTIBILITY_INFO,
		vmcs12->guest_interruptibility_info);
	kgr_vmcs_write32(GUEST_SYSENTER_CS, vmcs12->guest_sysenter_cs);
	kgr_kvm_set_dr(vcpu, 7, vmcs12->guest_dr7);
	kgr_vmx_set_rflags(vcpu, vmcs12->guest_rflags);
	kgr_vmcs_writel(GUEST_PENDING_DBG_EXCEPTIONS,
		vmcs12->guest_pending_dbg_exceptions);
	kgr_vmcs_writel(GUEST_SYSENTER_ESP, vmcs12->guest_sysenter_esp);
	kgr_vmcs_writel(GUEST_SYSENTER_EIP, vmcs12->guest_sysenter_eip);

	kgr_vmcs_write64(VMCS_LINK_POINTER, -1ull);

	exec_control = vmcs12->pin_based_vm_exec_control;
	exec_control |= kgr_vmcs_config->pin_based_exec_ctrl;
	exec_control &= ~(PIN_BASED_VMX_PREEMPTION_TIMER |
                          PIN_BASED_POSTED_INTR);
	kgr_vmcs_write32(PIN_BASED_VM_EXEC_CONTROL, exec_control);

	vmx->nested.preemption_timer_expired = false;
	if (kgr_nested_cpu_has_preemption_timer(vmcs12))
		kgr_vmx_start_preemption_timer(vcpu);

	/*
	 * Whether page-faults are trapped is determined by a combination of
	 * 3 settings: PFEC_MASK, PFEC_MATCH and EXCEPTION_BITMAP.PF.
	 * If enable_ept, L0 doesn't care about page faults and we should
	 * set all of these to L1's desires. However, if !enable_ept, L0 does
	 * care about (at least some) page faults, and because it is not easy
	 * (if at all possible?) to merge L0 and L1's desires, we simply ask
	 * to exit on each and every L2 page fault. This is done by setting
	 * MASK=MATCH=0 and (see below) EB.PF=1.
	 * Note that below we don't need special code to set EB.PF beyond the
	 * "or"ing of the EB of vmcs01 and vmcs12, because when enable_ept,
	 * vmcs01's EB.PF is 0 so the "or" will take vmcs12's value, and when
	 * !enable_ept, EB.PF is 1, so the "or" will always be 1.
	 *
	 * A problem with this approach (when !enable_ept) is that L1 may be
	 * injected with more page faults than it asked for. This could have
	 * caused problems, but in practice existing hypervisors don't care.
	 * To fix this, we will need to emulate the PFEC checking (on the L1
	 * page tables), using walk_addr(), when injecting PFs to L1.
	 */
	kgr_vmcs_write32(PAGE_FAULT_ERROR_CODE_MASK,
		*kgr_enable_ept ? vmcs12->page_fault_error_code_mask : 0);
	kgr_vmcs_write32(PAGE_FAULT_ERROR_CODE_MATCH,
		*kgr_enable_ept ? vmcs12->page_fault_error_code_match : 0);

	if (kgr_cpu_has_secondary_exec_ctrls()) {
		exec_control = kgr_vmx_secondary_exec_control(vmx);
		if (!vmx->rdtscp_enabled)
			exec_control &= ~SECONDARY_EXEC_RDTSCP;
		/* Take the following fields only from vmcs12 */
		exec_control &= ~(SECONDARY_EXEC_VIRTUALIZE_APIC_ACCESSES |
				  SECONDARY_EXEC_VIRTUAL_INTR_DELIVERY |
                                  SECONDARY_EXEC_APIC_REGISTER_VIRT);
		if (kgr_nested_cpu_has(vmcs12,
				CPU_BASED_ACTIVATE_SECONDARY_CONTROLS))
			exec_control |= vmcs12->secondary_vm_exec_control;

		if (exec_control & SECONDARY_EXEC_VIRTUALIZE_APIC_ACCESSES) {
			/*
			 * Translate L1 physical address to host physical
			 * address for vmcs02. Keep the page pinned, so this
			 * physical address remains valid. We keep a reference
			 * to it so we can release it later.
			 */
			if (vmx->nested.apic_access_page) /* shouldn't happen */
				kgr_nested_release_page(vmx->nested.apic_access_page);
			vmx->nested.apic_access_page =
				kgr_nested_get_page(vcpu, vmcs12->apic_access_addr);
			/*
			 * If translation failed, no matter: This feature asks
			 * to exit when accessing the given address, and if it
			 * can never be accessed, this feature won't do
			 * anything anyway.
			 */
			if (!vmx->nested.apic_access_page)
				exec_control &=
				  ~SECONDARY_EXEC_VIRTUALIZE_APIC_ACCESSES;
			else
				kgr_vmcs_write64(APIC_ACCESS_ADDR,
				  page_to_phys(vmx->nested.apic_access_page));
		} else if (kgr_vm_need_virtualize_apic_accesses(vmx->vcpu.kvm)) {
			exec_control |=
				SECONDARY_EXEC_VIRTUALIZE_APIC_ACCESSES;
			kgr_vmcs_write64(APIC_ACCESS_ADDR,
				page_to_phys(vcpu->kvm->arch.apic_access_page));
		}

		kgr_vmcs_write32(SECONDARY_VM_EXEC_CONTROL, exec_control);
	}


	/*
	 * Set host-state according to L0's settings (vmcs12 is irrelevant here)
	 * Some constant fields are set here by vmx_set_constant_host_state().
	 * Other fields are different per CPU, and will be set later when
	 * vmx_vcpu_load() is called, and when vmx_save_host_state() is called.
	 */
	kgr_vmx_set_constant_host_state(vmx);

	/*
	 * HOST_RSP is normally set correctly in vmx_vcpu_run() just before
	 * entry, but only if the current (host) sp changed from the value
	 * we wrote last (vmx->host_rsp). This cache is no longer relevant
	 * if we switch vmcs, and rather than hold a separate cache per vmcs,
	 * here we just force the write to happen on entry.
	 */
	vmx->host_rsp = 0;

	exec_control = kgr_vmx_exec_control(vmx); /* L0's desires */
	exec_control &= ~CPU_BASED_VIRTUAL_INTR_PENDING;
	exec_control &= ~CPU_BASED_VIRTUAL_NMI_PENDING;
	exec_control &= ~CPU_BASED_TPR_SHADOW;
	exec_control |= vmcs12->cpu_based_vm_exec_control;
	/*
	 * Merging of IO and MSR bitmaps not currently supported.
	 * Rather, exit every time.
	 */
	exec_control &= ~CPU_BASED_USE_MSR_BITMAPS;
	exec_control &= ~CPU_BASED_USE_IO_BITMAPS;
	exec_control |= CPU_BASED_UNCOND_IO_EXITING;

	kgr_vmcs_write32(CPU_BASED_VM_EXEC_CONTROL, exec_control);

	/* EXCEPTION_BITMAP and CR0_GUEST_HOST_MASK should basically be the
	 * bitwise-or of what L1 wants to trap for L2, and what we want to
	 * trap. Note that CR0.TS also needs updating - we do this later.
	 */
	kgr_update_exception_bitmap(vcpu);
	vcpu->arch.cr0_guest_owned_bits &= ~vmcs12->cr0_guest_host_mask;
	kgr_vmcs_writel(CR0_GUEST_HOST_MASK, ~vcpu->arch.cr0_guest_owned_bits);

	/* L2->L1 exit controls are emulated - the hardware exit is to L0 so
	 * we should use its exit controls. Note that VM_EXIT_LOAD_IA32_EFER
	 * bits are further modified by vmx_set_efer() below.
	 */
	kgr_vmcs_write32(VM_EXIT_CONTROLS, kgr_vmcs_config->vmexit_ctrl);

	/* vmcs12's VM_ENTRY_LOAD_IA32_EFER and VM_ENTRY_IA32E_MODE are
	 * emulated by vmx_set_efer(), below.
	 */
	kgr_vm_entry_controls_init(vmx, 
		(vmcs12->vm_entry_controls & ~VM_ENTRY_LOAD_IA32_EFER &
			~VM_ENTRY_IA32E_MODE) |
		(kgr_vmcs_config->vmentry_ctrl & ~VM_ENTRY_IA32E_MODE));

	if (vmcs12->vm_entry_controls & VM_ENTRY_LOAD_IA32_PAT) {
		kgr_vmcs_write64(GUEST_IA32_PAT, vmcs12->guest_ia32_pat);
		vcpu->arch.pat = vmcs12->guest_ia32_pat;
	} else if (kgr_vmcs_config->vmentry_ctrl & VM_ENTRY_LOAD_IA32_PAT)
		kgr_vmcs_write64(GUEST_IA32_PAT, vmx->vcpu.arch.pat);


	kgr_set_cr4_guest_host_mask(vmx);

	if (vmcs12->cpu_based_vm_exec_control & CPU_BASED_USE_TSC_OFFSETING)
		kgr_vmcs_write64(TSC_OFFSET,
			vmx->nested.vmcs01_tsc_offset + vmcs12->tsc_offset);
	else
		kgr_vmcs_write64(TSC_OFFSET, vmx->nested.vmcs01_tsc_offset);

	if (*kgr_enable_vpid) {
		/*
		 * Trivially support vpid by letting L2s share their parent
		 * L1's vpid. TODO: move to a more elaborate solution, giving
		 * each L2 its own vpid and exposing the vpid feature to L1.
		 */
		kgr_vmcs_write16(VIRTUAL_PROCESSOR_ID, vmx->vpid);
		kgr_vmx_flush_tlb(vcpu);
	}

	if (kgr_nested_cpu_has_ept(vmcs12)) {
		kgr_kvm_mmu_unload(vcpu);
		kgr_nested_ept_init_mmu_context(vcpu);
	}

	if (vmcs12->vm_entry_controls & VM_ENTRY_LOAD_IA32_EFER)
		vcpu->arch.efer = vmcs12->guest_ia32_efer;
	else if (vmcs12->vm_entry_controls & VM_ENTRY_IA32E_MODE)
		vcpu->arch.efer |= (EFER_LMA | EFER_LME);
	else
		vcpu->arch.efer &= ~(EFER_LMA | EFER_LME);
	/* Note: modifies VM_ENTRY/EXIT_CONTROLS and GUEST/HOST_IA32_EFER */
	kgr_vmx_set_efer(vcpu, vcpu->arch.efer);

	/*
	 * This sets GUEST_CR0 to vmcs12->guest_cr0, with possibly a modified
	 * TS bit (for lazy fpu) and bits which we consider mandatory enabled.
	 * The CR0_READ_SHADOW is what L2 should have expected to read given
	 * the specifications by L1; It's not enough to take
	 * vmcs12->cr0_read_shadow because on our cr0_guest_host_mask we we
	 * have more bits than L1 expected.
	 */
	kgr_vmx_set_cr0(vcpu, vmcs12->guest_cr0);
	kgr_vmcs_writel(CR0_READ_SHADOW, kgr_nested_read_cr0(vmcs12));

	kgr_vmx_set_cr4(vcpu, vmcs12->guest_cr4);
	kgr_vmcs_writel(CR4_READ_SHADOW, kgr_nested_read_cr4(vmcs12));

	/* shadow page tables on either EPT or shadow page tables */
	kgr_kvm_set_cr3(vcpu, vmcs12->guest_cr3);
	kgr_kvm_mmu_reset_context(vcpu);

	if (!*kgr_enable_ept)
		vcpu->arch.walk_mmu->inject_page_fault = kgr_vmx_inject_page_fault_nested;

	/*
	 * L1 may access the L2's PDPTR, so save them to construct vmcs12
	 */
	if (*kgr_enable_ept) {
		kgr_vmcs_write64(GUEST_PDPTR0, vmcs12->guest_pdptr0);
		kgr_vmcs_write64(GUEST_PDPTR1, vmcs12->guest_pdptr1);
		kgr_vmcs_write64(GUEST_PDPTR2, vmcs12->guest_pdptr2);
		kgr_vmcs_write64(GUEST_PDPTR3, vmcs12->guest_pdptr3);
	}

	kgr_kvm_register_write(vcpu, VCPU_REGS_RSP, vmcs12->guest_rsp);
	kgr_kvm_register_write(vcpu, VCPU_REGS_RIP, vmcs12->guest_rip);
}

/* optimized */
static void kgr_nested_vmx_entry_failure(struct kvm_vcpu *vcpu,
			struct vmcs12 *vmcs12,
			u32 reason, unsigned long qualification)
{
	kgr_load_vmcs12_host_state(vcpu, vmcs12);
	vmcs12->vm_exit_reason = reason | VMX_EXIT_REASONS_FAILED_VMENTRY;
	vmcs12->exit_qualification = qualification;
	kgr_nested_vmx_succeed(vcpu);
	if (*kgr_enable_shadow_vmcs)
		kgr_to_vmx(vcpu)->nested.sync_shadow_vmcs = true;
}



/* patched */
void kgr_vmx_save_host_state(struct kvm_vcpu *vcpu)
{
	struct vcpu_vmx *vmx = kgr_to_vmx(vcpu);
	int i;

	/*
	 * Fix CVE-2018-3646
	 *  +3 lines
	 *
	 * Note: we can't rely on the kgr_enable_ept pointer being
	 * valid when executing in the context of kvm.ko. Hence we
	 * must either check it here or live with L1D flushes on
	 * !enable_ept hosts.
	 */
	if (!*kgr_enable_ept)
		kgr_get_and_clear_vcpu_unconfined(vcpu, GFP_ATOMIC);

	if (vmx->host_state.loaded)
		return;

	vmx->host_state.loaded = 1;
	/*
	 * Set host fs and gs selectors.  Unfortunately, 22.2.3 does not
	 * allow segment selectors with cpl > 0 or ti == 1.
	 */
	vmx->host_state.ldt_sel = kvm_read_ldt();
	vmx->host_state.gs_ldt_reload_needed = vmx->host_state.ldt_sel;
	savesegment(fs, vmx->host_state.fs_sel);
	if (!(vmx->host_state.fs_sel & 7)) {
		kgr_vmcs_write16(HOST_FS_SELECTOR, vmx->host_state.fs_sel);
		vmx->host_state.fs_reload_needed = 0;
	} else {
		kgr_vmcs_write16(HOST_FS_SELECTOR, 0);
		vmx->host_state.fs_reload_needed = 1;
	}
	savesegment(gs, vmx->host_state.gs_sel);
	if (!(vmx->host_state.gs_sel & 7))
		kgr_vmcs_write16(HOST_GS_SELECTOR, vmx->host_state.gs_sel);
	else {
		kgr_vmcs_write16(HOST_GS_SELECTOR, 0);
		vmx->host_state.gs_ldt_reload_needed = 1;
	}

#ifdef CONFIG_X86_64
	savesegment(ds, vmx->host_state.ds_sel);
	savesegment(es, vmx->host_state.es_sel);
#endif

#ifdef CONFIG_X86_64
	kgr_vmcs_writel(HOST_FS_BASE, read_msr(MSR_FS_BASE));
	kgr_vmcs_writel(HOST_GS_BASE, read_msr(MSR_GS_BASE));
#else
	vmcs_writel(HOST_FS_BASE, kgr_segment_base(vmx->host_state.fs_sel));
	vmcs_writel(HOST_GS_BASE, kgr_segment_base(vmx->host_state.gs_sel));
#endif

#ifdef CONFIG_X86_64
	rdmsrl(MSR_KERNEL_GS_BASE, vmx->msr_host_kernel_gs_base);
	if (kgr_is_long_mode(&vmx->vcpu))
		wrmsrl(MSR_KERNEL_GS_BASE, vmx->msr_guest_kernel_gs_base);
#endif
	for (i = 0; i < vmx->save_nmsrs; ++i)
		kgr_kvm_set_shared_msr(vmx->guest_msrs[i].index,
				       vmx->guest_msrs[i].data,
				       vmx->guest_msrs[i].mask);
}

/* patched */
void kgr_vmx_handle_external_intr(struct kvm_vcpu *vcpu)
{
	u32 exit_intr_info = kgr_vmcs_read32(VM_EXIT_INTR_INFO);

	/*
	 * If external interrupt exists, IF bit is set in rflags/eflags on the
	 * interrupt stack frame, and interrupt will be enabled on a return
	 * from interrupt handler.
	 */
	if ((exit_intr_info & (INTR_INFO_VALID_MASK | INTR_INFO_INTR_TYPE_MASK))
			== (INTR_INFO_VALID_MASK | INTR_TYPE_EXT_INTR)) {
		unsigned int vector;
		unsigned long entry;
		gate_desc *desc;
		struct vcpu_vmx *vmx = kgr_to_vmx(vcpu);
#ifdef CONFIG_X86_64
		unsigned long tmp;
#endif

		vector =  exit_intr_info & INTR_INFO_VECTOR_MASK;
		desc = (gate_desc *)vmx->host_idt_base + vector;
		entry = gate_offset(*desc);
		asm volatile(
#ifdef CONFIG_X86_64
			"mov %%" _ASM_SP ", %[sp]\n\t"
			"and $0xfffffffffffffff0, %%" _ASM_SP "\n\t"
			"push $%c[ss]\n\t"
			"push %[sp]\n\t"
#endif
			"pushf\n\t"
			"orl $0x200, (%%" _ASM_SP ")\n\t"
			__ASM_SIZE(push) " $%c[cs]\n\t"
			"call *%[entry]\n\t"
			:
#ifdef CONFIG_X86_64
			[sp]"=&r"(tmp)
#endif
			:
			[entry]"r"(entry),
			[ss]"i"(__KERNEL_DS),
			[cs]"i"(__KERNEL_CS)
			);

		/*
		 * Fix CVE-2018-3646
		 *  +1 line
		 */
		kgr_set_vcpu_unconfined(vcpu, GFP_ATOMIC);

	} else
		local_irq_enable();
}

/* patched */
int kgr_nested_vmx_run(struct kvm_vcpu *vcpu, bool launch)
{
	struct vmcs12 *vmcs12;
	struct vcpu_vmx *vmx = kgr_to_vmx(vcpu);
	int cpu;
	struct loaded_vmcs *vmcs02;
	bool ia32e;

	if (!kgr_nested_vmx_check_permission(vcpu) ||
	    !kgr_nested_vmx_check_vmcs12(vcpu))
		return 1;

	kgr_skip_emulated_instruction(vcpu);
	vmcs12 = kgr_get_vmcs12(vcpu);

	if (*kgr_enable_shadow_vmcs)
		kgr_copy_shadow_to_vmcs12(vmx);

	/*
	 * The nested entry process starts with enforcing various prerequisites
	 * on vmcs12 as required by the Intel SDM, and act appropriately when
	 * they fail: As the SDM explains, some conditions should cause the
	 * instruction to fail, while others will cause the instruction to seem
	 * to succeed, but return an EXIT_REASON_INVALID_STATE.
	 * To speed up the normal (success) code path, we should avoid checking
	 * for misconfigurations which will anyway be caught by the processor
	 * when using the merged vmcs02.
	 */
	if (vmcs12->launch_state == launch) {
		kgr_nested_vmx_failValid(vcpu,
			launch ? VMXERR_VMLAUNCH_NONCLEAR_VMCS
			       : VMXERR_VMRESUME_NONLAUNCHED_VMCS);
		return 1;
	}

	if (vmcs12->guest_activity_state != GUEST_ACTIVITY_ACTIVE &&
	    vmcs12->guest_activity_state != GUEST_ACTIVITY_HLT) {
		kgr_nested_vmx_failValid(vcpu, VMXERR_ENTRY_INVALID_CONTROL_FIELD);
		return 1;
	}

	if ((vmcs12->cpu_based_vm_exec_control & CPU_BASED_USE_MSR_BITMAPS) &&
			!IS_ALIGNED(vmcs12->msr_bitmap, PAGE_SIZE)) {
		/*TODO: Also verify bits beyond physical address width are 0*/
		kgr_nested_vmx_failValid(vcpu, VMXERR_ENTRY_INVALID_CONTROL_FIELD);
		return 1;
	}

	if (kgr_nested_cpu_has2(vmcs12, SECONDARY_EXEC_VIRTUALIZE_APIC_ACCESSES) &&
			!IS_ALIGNED(vmcs12->apic_access_addr, PAGE_SIZE)) {
		/*TODO: Also verify bits beyond physical address width are 0*/
		kgr_nested_vmx_failValid(vcpu, VMXERR_ENTRY_INVALID_CONTROL_FIELD);
		return 1;
	}

	if (vmcs12->vm_entry_msr_load_count > 0 ||
	    vmcs12->vm_exit_msr_load_count > 0 ||
	    vmcs12->vm_exit_msr_store_count > 0) {
		pr_warn_ratelimited("%s: VMCS MSR_{LOAD,STORE} unsupported\n",
				    __func__);
		kgr_nested_vmx_failValid(vcpu, VMXERR_ENTRY_INVALID_CONTROL_FIELD);
		return 1;
	}

	if (!kgr_vmx_control_verify(vmcs12->cpu_based_vm_exec_control,
	      *kgr_nested_vmx_procbased_ctls_low, *kgr_nested_vmx_procbased_ctls_high) ||
	    !kgr_vmx_control_verify(vmcs12->secondary_vm_exec_control,
	      *kgr_nested_vmx_secondary_ctls_low, *kgr_nested_vmx_secondary_ctls_high) ||
	    !kgr_vmx_control_verify(vmcs12->pin_based_vm_exec_control,
	      *kgr_nested_vmx_pinbased_ctls_low, *kgr_nested_vmx_pinbased_ctls_high) ||
	    !kgr_vmx_control_verify(vmcs12->vm_exit_controls,
	      *kgr_nested_vmx_exit_ctls_low, *kgr_nested_vmx_exit_ctls_high) ||
	    !kgr_vmx_control_verify(vmcs12->vm_entry_controls,
	      *kgr_nested_vmx_entry_ctls_low, *kgr_nested_vmx_entry_ctls_high))
	{
		kgr_nested_vmx_failValid(vcpu, VMXERR_ENTRY_INVALID_CONTROL_FIELD);
		return 1;
	}

	if (((vmcs12->host_cr0 & KGR_VMXON_CR0_ALWAYSON) != KGR_VMXON_CR0_ALWAYSON) ||
	    ((vmcs12->host_cr4 & KGR_VMXON_CR4_ALWAYSON) != KGR_VMXON_CR4_ALWAYSON)) {
		kgr_nested_vmx_failValid(vcpu,
			VMXERR_ENTRY_INVALID_HOST_STATE_FIELD);
		return 1;
	}

	if (!kgr_nested_cr0_valid(vmcs12, vmcs12->guest_cr0) ||
	    ((vmcs12->guest_cr4 & KGR_VMXON_CR4_ALWAYSON) != KGR_VMXON_CR4_ALWAYSON)) {
		kgr_nested_vmx_entry_failure(vcpu, vmcs12,
			EXIT_REASON_INVALID_STATE, ENTRY_FAIL_DEFAULT);
		return 1;
	}
	if (vmcs12->vmcs_link_pointer != -1ull) {
		kgr_nested_vmx_entry_failure(vcpu, vmcs12,
			EXIT_REASON_INVALID_STATE, ENTRY_FAIL_VMCS_LINK_PTR);
		return 1;
	}

	/*
	 * If the load IA32_EFER VM-entry control is 1, the following checks
	 * are performed on the field for the IA32_EFER MSR:
	 * - Bits reserved in the IA32_EFER MSR must be 0.
	 * - Bit 10 (corresponding to IA32_EFER.LMA) must equal the value of
	 *   the IA-32e mode guest VM-exit control. It must also be identical
	 *   to bit 8 (LME) if bit 31 in the CR0 field (corresponding to
	 *   CR0.PG) is 1.
	 */
	if (vmcs12->vm_entry_controls & VM_ENTRY_LOAD_IA32_EFER) {
		ia32e = (vmcs12->vm_entry_controls & VM_ENTRY_IA32E_MODE) != 0;
		if (!kgr_kvm_valid_efer(vcpu, vmcs12->guest_ia32_efer) ||
		    ia32e != !!(vmcs12->guest_ia32_efer & EFER_LMA) ||
		    ((vmcs12->guest_cr0 & X86_CR0_PG) &&
		     ia32e != !!(vmcs12->guest_ia32_efer & EFER_LME))) {
			kgr_nested_vmx_entry_failure(vcpu, vmcs12,
				EXIT_REASON_INVALID_STATE, ENTRY_FAIL_DEFAULT);
			return 1;
		}
	}

	/*
	 * If the load IA32_EFER VM-exit control is 1, bits reserved in the
	 * IA32_EFER MSR must be 0 in the field for that register. In addition,
	 * the values of the LMA and LME bits in the field must each be that of
	 * the host address-space size VM-exit control.
	 */
	if (vmcs12->vm_exit_controls & VM_EXIT_LOAD_IA32_EFER) {
		ia32e = (vmcs12->vm_exit_controls &
			 VM_EXIT_HOST_ADDR_SPACE_SIZE) != 0;
		if (!kgr_kvm_valid_efer(vcpu, vmcs12->host_ia32_efer) ||
		    ia32e != !!(vmcs12->host_ia32_efer & EFER_LMA) ||
		    ia32e != !!(vmcs12->host_ia32_efer & EFER_LME)) {
			kgr_nested_vmx_entry_failure(vcpu, vmcs12,
				EXIT_REASON_INVALID_STATE, ENTRY_FAIL_DEFAULT);
			return 1;
		}
	}

	/*
	 * We're finally done with prerequisite checking, and can start with
	 * the nested entry.
	 */

	vmcs02 = kgr_nested_get_current_vmcs02(vmx);
	if (!vmcs02)
		return -ENOMEM;

	kgr_enter_guest_mode(vcpu);

	vmx->nested.vmcs01_tsc_offset = kgr_vmcs_read64(TSC_OFFSET);

	cpu = get_cpu();
	vmx->loaded_vmcs = vmcs02;
	kgr_vmx_vcpu_put(vcpu);
	kgr_vmx_vcpu_load(vcpu, cpu);
	vcpu->cpu = cpu;
	put_cpu();

	kgr_vmx_segment_cache_clear(vmx);

	vmcs12->launch_state = 1;

	kgr_prepare_vmcs02(vcpu, vmcs12);

	/*
	 * Fix CVE-2018-3646
	 *  +3 lines
	 */
	/* Hide L1D cache contents from the nested guest. */
	kgr_set_vcpu_unconfined(vcpu, GFP_KERNEL);

	if (vmcs12->guest_activity_state == GUEST_ACTIVITY_HLT)
		return kgr_kvm_emulate_halt(vcpu);

	vmx->nested.nested_run_pending = 1;

	/*
	 * Note no nested_vmx_succeed or nested_vmx_fail here. At this point
	 * we are no longer running L1, and VMLAUNCH/VMRESUME has not yet
	 * returned as far as L1 is concerned. It will only return (and set
	 * the success flag) when L2 exits (see nested_vmx_vmexit()).
	 */
	return 1;
}



static int kgr_patch_bsc1099306_kvm_intel_kallsyms(void)
{
	unsigned long addr;
	int i;

	for (i = 0; i < ARRAY_SIZE(kgr_funcs); i++) {
		/* mod_find_symname would be nice, but it is not exported */
		addr = kallsyms_lookup_name(kgr_funcs[i].name);
		if (!addr) {
			pr_err("kgraft-patch: symbol %s not resolved\n",
				kgr_funcs[i].name);
			return -ENOENT;
		}

		*(kgr_funcs[i].addr) = (void *)addr;
	}

	return 0;
}

static int
kgr_patch_bsc1099306_kvm_intel_module_notify(struct notifier_block *nb,
					     unsigned long action, void *data)
{
	struct module *mod = data;
	int ret;

	if (action != MODULE_STATE_COMING || strcmp(mod->name, KGR_PATCHED_MODULE))
		return 0;

	ret = kgr_patch_bsc1099306_kvm_intel_kallsyms();
	WARN(ret, "kgraft-patch: delayed kallsyms lookup failed. System is broken and can crash.\n");

	return ret;
}

static struct notifier_block kgr_patch_bsc1099306_kvm_intel_module_nb = {
	.notifier_call = kgr_patch_bsc1099306_kvm_intel_module_notify,
	.priority = INT_MIN+1,
};

int __kgr_patch_bsc1099306_kvm_intel_init(void)
{
	int ret;

	mutex_lock(&module_mutex);
	if (find_module(KGR_PATCHED_MODULE)) {
		ret = kgr_patch_bsc1099306_kvm_intel_kallsyms();
		if (ret)
			goto out;
	}

	ret = register_module_notifier(&kgr_patch_bsc1099306_kvm_intel_module_nb);
out:
	mutex_unlock(&module_mutex);
	return ret;
}

void __kgr_patch_bsc1099306_kvm_intel_cleanup(void)
{
	unregister_module_notifier(&kgr_patch_bsc1099306_kvm_intel_module_nb);
}

#endif /* IS_ENABLED(CONFIG_X86_64) && IS_ENABLED(CONFIG_KVM) */
